# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import subprocess
from pathlib import Path
from Bio import SeqIO
from ete3 import NCBITaxa

import fdog.libs.zzz as general_fn
import fdog.libs.fasta as fasta_fn
import fdog.libs.blast as blast_fn
import fdog.libs.output as output_fn
import fdog.libs.tree as tree_fn


##### FUNCTIONS FOR DATA/INPUT PREPARATION #####

def parsing_paths(args):
    """ Getting path to hmm core set, coreTaxa_dir, searchTaxa_dir and annotation_dir"""
    (pathFile, outpath, hmmpath, corepath, searchpath, annopath) = args
    ### get fdog and data path
    data_path = ''
    fdog_path = os.path.realpath(__file__).replace('/libs/preparation.py','')
    pathconfigFile = fdog_path + '/bin/pathconfig.yml'
    if not os.path.exists(pathconfigFile):
        sys.exit(
            f'No pathconfig.txt found at {pathconfigFile}. Please run fdog.setup '
            + '(https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).')

    if pathFile:
        pathconfigFile = os.path.abspath(pathFile)

    cfg = general_fn.load_config(pathconfigFile)
    try:
        data_path = cfg['dataPath']
    except:
        data_path = os.getcwd()

    if hmmpath == '':
        hmmpath = outpath + '/core_orthologs'
        Path(hmmpath).mkdir(parents = True, exist_ok = True)

    if corepath == '':
        try:
            corepath = cfg['corepath']
        except:
            corepath = data_path + '/coreTaxa_dir'
        general_fn.check_file_exist(corepath)
    if searchpath == '':
        try:
            searchpath = cfg['searchpath']
        except:
            searchpath = data_path + '/searchTaxa_dir'
        general_fn.check_file_exist(searchpath)
    if annopath == '':
        try:
            annopath = cfg['annopath']
        except:
            annopath = data_path + '/annotation_dir'
        general_fn.check_file_exist(annopath)
    return(hmmpath, corepath, searchpath, annopath)


def check_input(args):
    (seqFile, refspec, outpath, hmmpath, corepath,
        searchpath, annopath, pathFile) = args
    fdog_path = os.path.realpath(__file__).replace('/libs/preparation.py','')
    # create output directory
    Path(outpath).mkdir(parents = True, exist_ok = True)
    Path(hmmpath).mkdir(parents = True, exist_ok = True)
    # check path existing
    hmmpath, corepath, searchpath, annopath = parsing_paths(
        [pathFile, outpath, hmmpath, corepath, searchpath, annopath])
    for path in [hmmpath, corepath, searchpath, annopath]:
        general_fn.check_file_exist(path)
    # check for seqFile
    if not os.path.exists(os.path.abspath(seqFile)):
        if not os.path.exists(fdog_path + '/data/' + seqFile):
            sys.exit(
                'ERROR: %s not found in %s or %s'
                % (seqFile, os.getcwd(), fdog_path + '/data/'))
        else:
            seqFile = fdog_path + '/data/' + seqFile
    else:
        seqFile = os.path.abspath(seqFile)
    # check refspec
    if not os.path.exists(os.path.abspath(corepath+'/'+refspec)):
        exit('ERROR: Reference taxon %s not found in %s' % (refspec, corepath))
    return (seqFile, hmmpath, corepath, searchpath, annopath)


def check_blast_version(corepath, refspec):
    """ Check if blast DBs in corepath is compatible with blastp version """
    fdog_path = os.path.realpath(__file__).replace('/libs/preparation.py', '')
    query = os.path.join(fdog_path, 'data', 'infile.fa')
    blast_db = os.path.join(corepath, refspec, refspec)
    try:
        cmd = ["blastp", "-query", query, "-db", blast_db]
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
    except subprocess.CalledProcessError as e:
        sys.exit(f"ERROR: Error running BLAST (probably conflict with BLAST DB versions)\n{e.stderr}")


def check_ranks_core_taxa(corepath, refspec, minDist, maxDist):
    """ Check if refspec (or all core taxa) have a valid minDist and maxDist tax ID
    Return 2 dictionaries of taxa for invalid minDist and maxDist, where
    keys is taxon name and value is the next valid rank
    """
    invalid_minDist = []
    invalid_maxDist = []
    ncbi = NCBITaxa()
    rank_list = ['species', 'genus', 'family', 'order', 'class', 'phylum', 'kingdom', 'superkingdom']
    suggest_minIndex = rank_list.index(minDist)
    suggest_maxIndex = rank_list.index(maxDist)
    for f in os.listdir(corepath):
        if f == refspec: # remove this if need to check for all core taxa
            if os.path.isdir(f'{corepath}/{f}'):
                id = f.split('@')[1]
                lineage = ncbi.get_lineage(id)
                ranks = ncbi.get_rank(lineage)
                if len(general_fn.matching_elements(ranks, minDist)) < 1:
                    invalid_minDist.append(f)
                    index_minDist = rank_list.index(minDist) + 1
                    while index_minDist < len(rank_list):
                        if len(general_fn.matching_elements(ranks, rank_list[index_minDist])) > 0:
                            if index_minDist > suggest_minIndex:
                                suggest_minIndex = index_minDist
                            break
                        index_minDist += 1
                if len(general_fn.matching_elements(ranks, maxDist)) < 1:
                    invalid_maxDist.append(f)
                    index_maxDist = rank_list.index(maxDist) + 1
                    while index_maxDist < len(rank_list):
                        if len(general_fn.matching_elements(ranks, rank_list[index_maxDist])) > 0:
                            if index_maxDist > suggest_maxIndex:
                                suggest_maxIndex = index_maxDist
                            break
                        index_maxDist += 1
    return(invalid_minDist, invalid_maxDist, rank_list[suggest_minIndex], rank_list[suggest_maxIndex])


def get_seed_id_from_fa(core_fa, refspec):
    """ Get seed ID from core ortholog fasta file
    (used if --reuseCore option is specified)
    """
    core_seqs = SeqIO.to_dict((SeqIO.parse(open(core_fa), 'fasta')))
    core_ids = core_seqs.keys()
    seed_id = [s for s in core_ids if refspec in s][0].split('|')[-1]
    return(seed_id)


def identify_seed_id(seqFile, refspec, corepath, debug, silentOff):
    """ Identify seed ID in reference protein set using BLAST
    If the header of the seed fasta seq if found in the refspec proteome, then
    if can be directly used. Otherwise do blast search. If the return blast hit
    is longer/shorter than the seed sequence by 10 amino acids, fDOG will stop
    """
    refspec_db = '%s/%s/%s' % (corepath, refspec, refspec)
    # first check if input seed ID existiert in refspec genome
    refspec_fa = fasta_fn.read_fasta('%s.fa' % refspec_db)
    seed_fa = SeqIO.parse(open(seqFile),'fasta')
    for seed in seed_fa:
        try:
            if len(refspec_fa.fetch(seed.id)) == len(seed.seq):
                return(seed.id)
        except:
            output_fn.print_debug(debug, 'Identify seed ID', 'Input seed ID not found!')
    # otherwise, perform blast search
    blast_xml = blast_fn.do_blastsearch(seqFile, refspec_db, evalBlast = 0.001)
    if not blast_xml:
        print(f'ERROR: No blast output!')
        print(f'You can check it by running:\nblastp -query {seqFile} -db {corepath}/{refspec}/{refspec} -evalue 0.001 -outfmt 7')
        sys.exit()
    blast_out = blast_fn.parse_blast_xml(blast_xml)
    if len(blast_out['hits']) < 1:
        print(f'ERROR: Cannot find seed sequence {blast_out["query"]} in genome of reference species (No blast hit was found)!')
        print(f'You can check it by running:\nblastp -query {seqFile} -db {corepath}/{refspec}/{refspec} -evalue 0.001 -outfmt 7')
        sys.exit()
    for hit in blast_out['hits']:
        if blast_out['hits'][hit]['align_len'] == blast_out['query_len']:
            return(hit)
        elif abs(int(blast_out['hits'][hit]['align_len']) - int(blast_out['query_len'])) < 10:
            output_fn.print_stdout(silentOff, 'WARNING: Found seed sequence shorter/longer than input!')
            return(hit)
        else:
            print(f'ERROR: Cannot find seed sequence {blast_out["query"]} in genome of reference species (blast hit is much longer/shorter than seed)!')
            print(f'You can check it by running:\nblastp -query {seqFile} -db {corepath}/{refspec}/{refspec} -evalue 0.001 -outfmt 7')
            sys.exit()



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
from pathlib import Path
import ssl
import urllib.request
import yaml
import time
import pickle


##### GENERAL FUNCTIONS FOR FILES, FOLDERS AND GENERAL VARIABLES #####

def check_file_exist(file):
    """ Exit if a file does not exist"""
    if not os.path.exists(os.path.abspath(file)):
        sys.exit('%s not found' % file)


def read_file(file):
    """ Read a file and return list of lines"""
    if os.path.exists(file):
        with open(file, 'r') as f:
            lines = f.read().splitlines()
            f.close()
            return(lines)
    else:
        sys.exit('%s not found' % file)


def read_dir(dir):
    """ Return list of directories from a given path """
    check_file_exist(dir)
    out_dirs = []
    p = os.listdir(dir)
    for i in p:
        if os.path.isdir('%s/%s' % (dir, i)):
            out_dirs.append(i)
    return(out_dirs)


def load_config(config_file):
    """ Load a YAML file and return as a dictionary """
    with open(config_file, 'r') as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)


def download_progress(count, block_size, total_size):
    global start_time
    if count == 0:
        start_time = time.time()
        return
    duration = time.time() - start_time
    progress_size = int(count * block_size)
    speed = int(progress_size / (1024 * duration))
    percent = int(count * block_size * 100 / total_size)
    if percent > 100:
        percent = 100
    sys.stdout.write("\r...%d%%, %d MB, %d KB/s, %d seconds passed" %
                     (percent, progress_size / (1024 * 1024), speed, duration))
    sys.stdout.flush()


def download_file(url, file):
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    download_file = urllib.request.URLopener(context=ctx)
    print('Downloading %s' % (url + '/' + file))
    urllib.request.urlretrieve(url + '/' + file, file, download_progress)
    print(' ... done!')


def count_line(file, pattern, contain):
    """ Count lines in file that contain (or not) a pattern """
    nline = 0
    with open(file, 'r') as f:
        for line in f:
            if contain:
                if pattern in line:
                    nline = nline + 1
            else:
                if not pattern in line:
                    nline = nline + 1
    return(nline)


def get_ids_from_folder(folder, type):
    """ Get taxonomy IDs for from coreTaxa_dir, searchTaxa_dir or annotation_dir
    Return dictionary {taxID:<TaxName>@<TaxID>@Ver}
    """
    tax_ids = {}

    for name in read_dir(folder):
        if type == 'annotation_dir':
            if not name.endswith('.json'):
                continue
            else:
                name = name.replace('.json','')
        else:
            if not os.path.isdir('%s/%s' % (folder, name)):
                continue
        id = name.split('@')[1]
        if not id in tax_ids:
            tax_ids[id] = name
    return(tax_ids)


def matching_elements(dictionary, search_string):
    """ Search for a string in dictionary's values
    Return {key:val} where string was found in val
    """
    return {key:val for key,val in dictionary.items() if search_string == val}


def remove_dup_in_dict(dictionary):
    """ Find and remove duplicated or empty values of a dictionary """
    tmp_dict = {'_'.join(val) : key for key, val in dictionary.items()}
    res = {val : key.split('_') for key, val in tmp_dict.items()}
    res = {key : val for key, val in res.items() if len(val[0]) > 0}
    return(res)


def join_2lists(first_list, second_list):
    """ Join 2 lists """
    in_first = set(first_list)
    in_second = set(second_list)
    in_second_but_not_in_first = in_second - in_first
    out = first_list + list(in_second_but_not_in_first)
    return(out)


def save_pyobj(obj, out_file):
    """ Save a python object to out_file """
    with open(out_file, 'wb') as obj_out:
        pickle.dump(obj, obj_out)


def read_pyobj_file(in_file):
    """ Read a python object from an in_file """
    with open(in_file, 'rb') as obj_file:
        return(pickle.load(obj_file))


def query_yes_no(question, default='yes'):
    valid = {'yes': True, 'y': True, 'ye': True,
             'no': False, 'n': False}
    if default is None:
        prompt = ' [y/n] '
    elif default == 'yes':
        prompt = ' [Y/n] '
    elif default == 'no':
        prompt = ' [y/N] '
    else:
        raise ValueError('invalid default answer: "%s"' % default)
    while True:
        choice = sys.stdin.readline().rstrip().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write('Please respond with "yes" or "no" '
                             '(or "y" or "n").\n')


def search_string_in_file(file, string):
    """ Search for a string in file
    Return 0 if not found, 1 if found
    """
    flag = 0
    with open(file, 'r') as fp:
        for l_no, line in enumerate(fp):
            if string in line:
                flag = 1
                break
    return(flag)



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import subprocess
import shutil
import greedyFAS.annoFAS.annoModules as annoFas

import fdog.libs.zzz as general_fn


##### FUNCTIONS RELATED TO FAS #####

def check_fas_executable():
    try:
        subprocess.check_output(['fas.setup -t ./ --check'], shell = True, stderr = subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        print('\033[96m%s\033[0m' % e.output.decode(sys.stdout.encoding).strip())
        print('FAS installed but fas.setup still need to be run if you want to use it!')
        return(0)
    return(1)


def get_tool_fas_path():
    """ Get path to FAS annotation tools """
    cmd = 'fas.setup -t ~/ -c'
    try:
        out = subprocess.run(
                [cmd], shell = True, capture_output = True, check = True)
        tool_path = out.stdout.decode().split('\n')[0].split()[6].replace('.','')
        return(tool_path)
    except:
        sys.exit('ERROR: fas.setup cannot be called!')


def get_anno_fas(seqName, spec, seq_id, seq, hmmpath, annopath):
    """ Get annotation for a seq_id from existing json file in annopath """
    out_json = '%s/%s/%s_%s.json' % (hmmpath, seqName, spec, seq_id)
    if not os.path.exists(out_json):
        tmp_seed_fa = '%s/%s/%s_%s.fa' % (hmmpath, seqName, seqName, spec)
        with open(tmp_seed_fa, 'w') as tmp_seed_fa_out:
            tmp_seed_fa_out.write('>%s\n%s\n' % (seq_id, seq))
        spec_anno = '%s/%s.json' % (annopath, spec)
        try:
            anno_dict = annoFas.extractAnno(tmp_seed_fa, spec_anno)
            anno_dict['clan'] = annoFas.getClans(
                get_tool_fas_path(), anno_dict['feature'])
            annoFas.save2json(
                anno_dict, '%s_%s' % (spec, seq_id), '%s/%s' % (hmmpath, seqName))
            os.remove(tmp_seed_fa)
        except:
            sys.exit(
                'ERROR: Annotation for %s cannot be found in %s'
                % (seq_id, spec_anno))
    return(out_json)


def calc_pairwise_fas(seed_json, query_json, seqName, hmmpath):
    """ Calculate FAS score for a pair seed and query protein
    Input are two anno json files for seed and query
    Return a value between 0 and 1
    """
    general_fn.check_file_exist(seed_json)
    general_fn.check_file_exist(query_json)

    fas_cmd = 'fas.run -s %s -q %s --no_config' % (seed_json, query_json)
    fas_cmd = '%s -a %s/%s --raw --tsv --domain --cpus 1 -o %s/%s' \
                % (fas_cmd, hmmpath, seqName, hmmpath, seqName)
    try:
        fas_out = subprocess.run(
                [fas_cmd], shell = True, capture_output = True, check = True)
    except:
        sys.exit('ERROR: Error running FAS\n%s' % fas_cmd)
    results = fas_out.stdout.decode().split('\n')
    for l in results:
        if l.startswith('#') and len(l.split('\t')) > 1:
            return(l.split('\t')[-1])
    return('')


def calc_fas_cand(args):
    """ Calculate FAS score for a ortholog candidate against seed
    Ortholog candidate defined by spec, seq_id and seq
    """
    (fasOff, seqName, seed_json, spec, seq_id, seq, hmmpath, annopath) = args
    if not fasOff == True:
        query_json = get_anno_fas(seqName, spec, seq_id, seq, hmmpath, annopath)
        fas_score = calc_pairwise_fas(seed_json, query_json, seqName, hmmpath)
    else:
        fas_score = 1
    return(fas_score)


def calc_fas_multi (input_fa, outpath, annopath, cpus, featureFile):
    """ Calculate pairwise FAS scores for all orthologs vs seed protein
    input_fa is the default <seqName>.extended.fa output file of fDOG
    Output will be <seqName>_forward.
    """
    fasCmd = 'fas.runFdogFas -i %s -w %s --cores %s --redo_anno' % (input_fa, annopath, cpus)
    if featureFile:
        fasCmd = f'{fasCmd} --featuretypes {featureFile}'
    try:
        subprocess.call([fasCmd], shell = True)
        if os.path.exists(outpath + '/tmp'):
            shutil.rmtree(outpath + '/tmp')
    except:
        sys.exit('Problem running\n%s' % (fasCmd))



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import os
from Bio import SeqIO
import multiprocessing as mp
from tqdm import tqdm
import time

import fdog.libs.zzz as general_fn
import fdog.libs.fasta as fasta_fn
import fdog.libs.blast as blast_fn
import fdog.libs.hmm as hmm_fn
import fdog.libs.alignment as align_fn
import fdog.libs.preparation as prepare_fn
import fdog.libs.output as output_fn


##### FUNCTION FOR HMM-BASED ORTHOLOG SEARCH (HaMStR) #####
def hamstr(args):
    (seqName, hmmpath, corepath, searchpath, outpath,
        refspec, seed_id, search_taxon,
        evalHmmer, hitLimit, hmmScoreType, scoreCutoff,
        evalBlast, lowComplexityFilter,
        checkCoorthologsRefOff, rbh, rep,
        aligner, cpus, debug, silentOff, noCleanup) = args
    """ Ortholog search algorithm for a hmm core group agains a search taxon
    Implemented based on HaMStR https://doi.org/10.1186/1471-2148-9-157
    """
    ### (0) Dict for storing candidate and final orthologs (key=id, value=seq)
    ortho_candi = {}
    ortho_final = {}
    ### (00) Parse input files
    hmm_file = '%s/%s/hmm_dir/%s.hmm' % (hmmpath, seqName, seqName)
    refspec_db = '%s/%s/%s' % (corepath, refspec, refspec)
    refspec_fa = '%s/%s/%s.fa' % (corepath, refspec, refspec)
    search_fa = '%s/%s/%s.fa' % (searchpath, search_taxon, search_taxon)
    ### (000) Adapt parameters
    if rbh == True:
        checkCoorthologsRefOff = True
        rep = True

    ### PRINT JOB PARAMETERS
    output_fn.print_stdout(
        silentOff,
        '\n### Ortholog search ###'
        + '\nSeed: %s\nRefspec: %s\n' % (seqName, refspec)
        + 'Ref_seqID: %s\n' % seed_id
        + 'Search taxon: %s' % search_taxon)
    output_fn.print_debug(
        debug, 'Parameters',
        'HMM evalue cutoff: %s\nHMM hit limit: %s\n' % (evalHmmer, hitLimit)
        + 'HMM hit score cutoff: %s\n' % scoreCutoff
        + 'BLAST evalue cutoff: %s\n' % evalBlast
        + 'Blast low complexity filter: %s\n' % lowComplexityFilter
        + 'Turn off check for co-orthologs ref: %s\n' % checkCoorthologsRefOff
        + 'Aligner: %s' % aligner)

    ### (1) Do hmmsearch for query hmm against search taxon fasta
    hmm_hits = hmm_fn.do_hmmsearch(
            hmm_file, search_fa, evalHmmer, scoreCutoff, hitLimit, hmmScoreType, cpus, debug)
    output_fn.print_debug(debug, 'Sorted HMM hits', hmm_hits)
    ### (2) Read fasta file of refspec and search taxon
    refspec_seqs = fasta_fn.read_fasta(refspec_fa)
    search_seqs = fasta_fn.read_fasta(search_fa)
    ### (3) Do re-blast search for each hmm hit against refspec
    if len(hmm_hits) == 0:
        output_fn.print_stdout(
            silentOff, 'WARNING: No HMM hit found!')
    else:
        for hmm_hit in hmm_hits:
            hmm_hit_fa = '%s/hmm_%s_%s_%s.fa' % (
                                outpath, seqName, search_taxon, hmm_hit)
            with open(hmm_hit_fa, 'w') as hmm_fa_out:
                hmm_fa_out.write('>%s_%s\n%s' % (search_taxon, hmm_hit, search_seqs.fetch(hmm_hit)))
            blast_xml = blast_fn.do_blastsearch(
                    hmm_hit_fa, refspec_db, evalBlast = evalBlast, lowComplexityFilter = lowComplexityFilter)
            blast_out = blast_fn.parse_blast_xml(blast_xml)
            output_fn.print_debug(debug, 'BLAST hits', blast_out)
            if noCleanup == False:
                os.remove(hmm_hit_fa)
            ### (4) check reciprocity
            ### (4a) if refspec_seq_id == best blast hit
            if len(blast_out['hits'].keys()) > 0:
                best_blast_hit = list(blast_out['hits'].keys())[0]
                if best_blast_hit == hmm_hit and len(blast_out['hits'].keys()) > 1:
                    best_blast_hit = list(blast_out['hits'].keys())[0]
                if seed_id == best_blast_hit:
                    output_fn.print_stdout(
                        silentOff,
                        '%s accepted (best blast hit is ref)' % (blast_out['query']))
                    ortho_candi[hmm_hit] = search_seqs.fetch(hmm_hit)
                    continue
                else:
                    ### (4b) else, check for co-ortholog ref
                    if checkCoorthologsRefOff == False:
                        aln_fa = '%s/blast_%s_%s_%s_%s_%s.fa' % (
                                    outpath, seqName, seed_id, search_taxon,
                                    hmm_hit, best_blast_hit)
                        with open(aln_fa, 'w') as aln_fa_out:
                            aln_fa_out.write(
                                '>%s\n%s\n>%s_%s\n%s\n>%s\n%s' % (
                                    seed_id, refspec_seqs.fetch(seed_id),
                                    search_taxon, hmm_hit, search_seqs.fetch(hmm_hit),
                                    best_blast_hit, refspec_seqs.fetch(best_blast_hit)
                                )
                            )
                        fasta_fn.remove_dup(aln_fa)
                        aln_seq = align_fn.do_align(aligner, aln_fa)
                        output_fn.print_debug(
                            debug, 'Alignment for checking co-ortholog ref', aln_seq)
                        br_dist = align_fn.calc_Kimura_dist(aln_seq, best_blast_hit, seed_id, debug)
                        bh_dist = align_fn.calc_Kimura_dist(aln_seq, best_blast_hit, f'{search_taxon}_{hmm_hit}', debug)
                        output_fn.print_debug(
                            debug, 'Check if distance blast_vs_ref < blast_vs_hmm',
                            'd_br = %s; d_bh = %s' % (br_dist, bh_dist))
                        if noCleanup == False:
                            os.remove(aln_fa)
                        if br_dist == bh_dist == 0 or br_dist < bh_dist:
                            output_fn.print_stdout(
                                silentOff,
                                '%s accepted (best blast hit is co-ortholog to ref)'
                                % (blast_out['query'])
                            )
                            ortho_candi[hmm_hit] = search_seqs.fetch(hmm_hit)
                            continue

        # remove seed protein from candidate list
        if search_taxon == refspec:
            if seed_id in ortho_candi:
                ortho_candi.pop(seed_id)
        ### (5) check co-ortholog if more than 1 HMM hits are accepted
        if len(ortho_candi) == 0:
            output_fn.print_stdout(
                silentOff, 'WARNING: Reciprocity not fulfulled! No ortholog found!')
        else:
            output_fn.print_debug(
                debug, 'Candidates for checking co-orthologs', ortho_candi.keys())
            best_ortho = list(ortho_candi.keys())[0]
            ortho_final = fasta_fn.add_seq_to_dict(
                ortho_final, '%s|%s|%s|1' % (seqName, search_taxon, best_ortho),
                ortho_candi[best_ortho])

            if rep == False:
                if len(ortho_candi) > 1:
                    aln_co_fa = '%s/coortho_%s_%s.fa' % (
                                outpath, seqName, search_taxon)
                    with open(aln_co_fa, 'w') as aln_co_fa_out:
                        aln_co_fa_out.write(('>%s\n%s\n') %
                            (seed_id, refspec_seqs.fetch(seed_id)))
                        for cand in ortho_candi:
                            aln_co_fa_out.write(('>%s\n%s\n') %
                                (f'{search_taxon}_{cand}', ortho_candi[cand]))
                    aln_co_seq = align_fn.do_align(aligner, aln_co_fa)
                    output_fn.print_debug(
                        debug, 'Alignment for checking co-orthologs', aln_co_seq)
                    if noCleanup == False:
                        os.remove(aln_co_fa)
                    best_dist = align_fn.calc_Kimura_dist(
                                    aln_co_seq, seed_id, f'{search_taxon}_{best_ortho}', debug)
                    for cand in ortho_candi:
                        if not cand == best_ortho:
                            candi_dist = align_fn.calc_Kimura_dist(
                                aln_co_seq, f'{search_taxon}_{best_ortho}', f'{search_taxon}_{cand}', debug)
                            output_fn.print_debug(
                                debug,
                                'Check if distance bestHmm_vs_ref > '
                                + 'other_vs_bestHmm',
                                'd_best = %s; d_other = %s'
                                % (best_dist, candi_dist))
                            if candi_dist < best_dist:
                                if not cand == seed_id:
                                    ortho_final = fasta_fn.add_seq_to_dict(
                                        ortho_final,
                                        '%s|%s|%s|0' \
                                            % (seqName, search_taxon, cand),
                                        ortho_candi[cand])
            output_fn.print_stdout(
                silentOff,
                '=> %s orthologs found: %s'
                % (len(ortho_final), list(ortho_final.keys())))
    return(ortho_final)


def run_hamstr(args):
    """ Perform ortholog search based on hamstr approach """

    (seqName, refspec, pathArgs, orthoArgs, otherArgs) = args
    (outpath, hmmpath, corepath, searchpath, annopath) = pathArgs
    (checkCoorthologsRefOff, rbh, rep, evalBlast, lowComplexityFilter,
                        evalHmmer, hitLimit, hmmScoreType, scoreCutoff, aligner) = orthoArgs
    (searchTaxa, cpus, debug, silentOff, noCleanup, force, append) = otherArgs

    hamstr_jobs = []
    ### get ref seqID
    core_fa = '%s/%s/%s.fa' % (hmmpath, seqName, seqName)
    seed_id = prepare_fn.get_seed_id_from_fa(core_fa, refspec)

    ### get search taxa from user defined list (as a file or directly a list)
    if not searchTaxa == '':
        ignored_taxa = []
        if os.path.exists(os.path.abspath(searchTaxa)):
            search_taxa = general_fn.read_file(searchTaxa)
        else:
            search_taxa = searchTaxa.split(',')

        for search_taxon in search_taxa:
            if os.path.exists(
                    os.path.abspath(
                        '%s/%s/%s.fa' % (searchpath,search_taxon,search_taxon))):
                hamstr_jobs.append([
                    seqName, hmmpath, corepath, searchpath, outpath,
                    refspec, seed_id, search_taxon,
                    evalHmmer, hitLimit, hmmScoreType, scoreCutoff,
                    evalBlast, lowComplexityFilter,
                    checkCoorthologsRefOff, rbh, rep,
                    aligner, cpus, debug, silentOff, noCleanup
                ])
            else:
                ignored_taxa.append(search_taxon)
        if len(ignored_taxa) > 0:
            print(
                'WARNING: %s taxa cannot be found at %s\n%s'
                % (len(ignored_taxa), searchpath, ignored_taxa))
    ### get search taxa from searchpath (searchTaxa_dir)
    else:
        for search_taxon in general_fn.read_dir(searchpath):
            if os.path.exists(
                    os.path.abspath(
                        '%s/%s/%s.fa' % (searchpath,search_taxon,search_taxon))):
                hamstr_jobs.append([
                    seqName, hmmpath, corepath, searchpath, outpath,
                    refspec, seed_id, search_taxon,
                    evalHmmer, hitLimit, hmmScoreType, scoreCutoff,
                    evalBlast, lowComplexityFilter,
                    checkCoorthologsRefOff, rbh, rep,
                    aligner, cpus, debug, silentOff, noCleanup
                ])

    ### do ortholog search
    hamstr_out = {}
    if len(hamstr_jobs) > 0:
        if cpus > len(hamstr_jobs):
            cpus = len(hamstr_jobs)
        output_fn.print_stdout(
                silentOff, 'Ortholog search for %s taxa...' % len(hamstr_jobs))
        if debug == True or silentOff == True or len(hamstr_jobs) == 1:
            for job in hamstr_jobs:
                tmp_out = hamstr(job)
                hamstr_out = {**hamstr_out, **tmp_out}
        else:
            pool = mp.Pool(cpus)
            for _ in tqdm(
                    pool.imap_unordered(hamstr, hamstr_jobs),
                    total=len(hamstr_jobs)):
                if len(_) > 0:
                    hamstr_out = {**hamstr_out, **_}

    ### Get seed seq
    refspec_fa = '%s/%s/%s.fa' % (corepath, refspec, refspec)
    refspec_seqs = fasta_fn.read_fasta(refspec_fa)
    seed_id_mod = '%s|%s|%s|1' % (seqName, refspec, seed_id)
    seed_seq = refspec_seqs.fetch(seed_id)

    ### return
    return({**{seed_id_mod:seed_seq}, **hamstr_out})


def get_search_taxa_ids(searchTaxa, searchpath):
    """ Get taxonomy IDs for search taxa
    Either from searchTaxa_dir, or from user input list (--searchTaxa)
    Return dictionary {taxID:<TaxName>@<TaxID>@Ver}
    """
    tax_ids = {}
    if not searchTaxa == '' and not len(searchTaxa) == 0:
        ignored_taxa = []
        if os.path.exists(os.path.abspath(searchTaxa)):
            search_taxa = general_fn.read_file(searchTaxa)
        else:
            search_taxa = searchTaxa.split(',')

        for search_taxon in search_taxa:
            if not os.path.exists(
                    os.path.abspath(
                        '%s/%s/%s.fa' % (searchpath,search_taxon,search_taxon))):
                ignored_taxa.append(search_taxon)
            else:
                id = search_taxon.split('@')[1]
                if not id in tax_ids:
                    tax_ids[id] = search_taxon
        if len(ignored_taxa) > 0:
            print(
                'WARNING: %s taxa cannot be found at %s\n%s'
                % (len(ignored_taxa), searchpath, ignored_taxa))
    else:
        tax_ids = general_fn.get_ids_from_folder(searchpath, 'searchTaxa_dir')
    return(tax_ids)



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import subprocess
import math
import re
from Bio import SeqIO
from io import StringIO
import random

import fdog.libs.fasta as fasta_fn
import fdog.libs.output as output_fn

##### FUNCTIONS RELATED TO SEQ ALIGNMENT #####

def check_fasta36_executable(fdogPath):
    """ Check if FASTA36 installed in fdogPath """
    try:
        fasta36_cmd = '%s/bin/aligner/bin/ggsearch36' % fdogPath
        subprocess.check_output(fasta36_cmd, shell = True, stderr = subprocess.STDOUT)
        return('%s/bin/aligner/bin/' % fdogPath)
    except:
        try:
            which_fasta36 = subprocess.run(
                'which fasta36', shell = True, capture_output = True, check = True)
            return(which_fasta36.stdout.decode().strip().replace('fasta36',''))
        except subprocess.CalledProcessError as e:
            sys.exit('\033[91mERROR: FASTA36 not found!\033[0m')


def get_muscle_version(aligner):
    """ Check muscle version (3.8 or 5.1)
    Return v3 for v3.8, otherwise v5
    """
    cmd = 'muscle -version'
    try:
        out = subprocess.run(cmd, shell = True, capture_output = True, check = True)
        if 'v3.8' in out.stdout.decode():
            return('v3')
        else:
            return('v5')
    except subprocess.CalledProcessError as e:
        raise RuntimeError("Command '{}' return with error (code {}): {}".format(e.cmd, e.returncode, e.output))
        sys.exit('\033[91mERROR: Error running command\n%s\033[0m' % cmd)


def do_align(aligner, fa_file):
    """ Do alignment using MUSCLE or MAFFT for a multiple fasta file
    Return a dictionary (SeqIO object) containing seq IDs and aligned sequences
    Note: if any input seq is longer than 12.000 aa/nt, only MAFFT can be used
    """
    input_fa = SeqIO.to_dict((SeqIO.parse(open(fa_file), 'fasta')))
    if len(input_fa) == 1:
        return(input_fa)
    # parse output file name (otherwise cause troubles for muscle_v5)
    out_file = fa_file.split('/')[-1].replace('@', '_')
    # check muscle version
    if aligner == 'muscle':
        if get_muscle_version(aligner) == 'v3':
            if fasta_fn.check_long_seq(fa_file, 12000) == 1:
                aligner = 'mafft-linsi'
            else:
                aligner = 'muscle_v3'
        else:
            if fasta_fn.check_long_seq(fa_file, 15000) == 1:
                aligner = 'mafft-linsi'
            else:
                aligner = 'muscle_v5'
    # create alignment command and run
    align_cline = ''
    if aligner == 'muscle_v3':
        align_cline = 'muscle -in %s' % fa_file
    elif aligner == 'muscle_v5':
        align_cline = 'muscle -align %s -output %s.muscle.out' % (fa_file, out_file)
    else:
        align_cline = 'mafft --localpair --maxiterate 1000 %s' % fa_file
    try:
        aln_out = subprocess.run([align_cline], shell = True, capture_output = True, check = True)
    except subprocess.CalledProcessError as e:
        raise RuntimeError("Command '{}' return with error (code {}): {}".format(e.cmd, e.returncode, e.output))
        sys.exit(
            'ERROR: Error doing alignment with %s for %s' % (aligner, fa_file))

    if aligner == 'muscle_v5':
        aln_seq = SeqIO.to_dict((SeqIO.parse(open('%s.muscle.out' % out_file), 'fasta')))
        os.remove('%s.muscle.out' % out_file)
    else:
        aln_io = StringIO(aln_out.stdout.decode().strip())
        aln_seq = SeqIO.to_dict((SeqIO.parse(aln_io,'fasta')))
    return(aln_seq)


def calc_Kimura_dist(aln_dict, id_1, id_2, debug):
    """ Calculate Kimura distance for a pair of sequences
    Input is a dictionary of MSA (see do_align function).
    The Kimura distance is calculated based on perl module
    https://metacpan.org/pod/Bio::Align::ProteinStatistics#D-distance-methods
    """
    matches = 0
    total = 0
    if id_1 in aln_dict and id_2 in aln_dict:
        for a, b in zip(aln_dict[id_1].seq, aln_dict[id_2].seq):
            if a != '-' and b != '-':
                if a == b:
                    matches +=1
                total += 1
        if not total == 0:
            D = 1 - (matches/total)
        else:
            D = 1
        output_fn.print_debug(
            debug, 'Kimura distance',
            'kimura = round(- (math.log( 1 - %s - (0.2 * (%s ** 2)))), 5)' % (D, D))
        try:
            kimura = round(- (math.log( 1 - D - (0.2 * (D ** 2)))), 5)
        except:
            kimura = 999
        return(kimura)
    else:
        sys.exit('ERROR: %s or %s not found in %s!' % (id_1, id_2, aln_dict))


def calc_aln_score(fa1, fa2, aln_strategy = 'local', debugCore = False):
    """ Calculate alignment score for genes in fa2 vs other genes in fa1
    Return dictionary {gene_id:aln_score}
    """
    fdog_path = os.path.realpath(__file__).replace('/libs/alignment.py','')
    fa1_filename = fa1.split("/")[-1]
    fa2_filename = fa2.split("/")[-1]
    if os.path.exists(fa1_filename):
        if fa2_filename == fa1_filename:
            fa2_filename = f'{fa2_filename}.core'
        fa1_filename = f'{fa1_filename}.core'
    os.symlink(fa1, fa1_filename)
    if not fa2_filename == fa1_filename:
        os.symlink(fa2, fa2_filename)
    fasta36_options = f'{fa1_filename} {fa2_filename} -s BP62 -m 9 -d 0 -z -1 -E 100'
    fdog_path = os.path.realpath(__file__).replace('/libs/alignment.py','')
    fasta36_bin = check_fasta36_executable(fdog_path)
    if aln_strategy == 'global':
        fasta36_cmd = '%s/ggsearch36 %s' \
                                % (fasta36_bin, fasta36_options)
    elif aln_strategy == 'glocal':
        fasta36_cmd = '%s/glsearch36 %s' \
                                % (fasta36_bin, fasta36_options)
    else:
        fasta36_cmd = '%s/ssearch36 %s' \
                                % (fasta36_bin, fasta36_options)
    output_fn.print_debug(
        debugCore, 'ALN SCORE',
        'Calculate aln score using FASTA36: %s' % fasta36_cmd)
    try:
        fasta36_out = subprocess.run(
                [fasta36_cmd], shell = True, capture_output = True, check = True)
    except:
        sys.exit('ERROR: Error running FASTA36\n%s' % fasta36_cmd)
    # returns score for genes in fa2
    aln_score = {}
    cand_dict = SeqIO.to_dict((SeqIO.parse(open(fa2), 'fasta')))
    for id in list(cand_dict.keys()):
        aln_score[id[0:60]] = 0
    results = fasta36_out.stdout.decode().split('\n')
    for l in results:
        if len(l) > 1:
            gene_id = l.split()[0]
            if gene_id in aln_score:
                if re.search('\(\s+\d+\)', l):
                    l = re.sub(r'\(\s+','(', l)
                aln_score[gene_id] = aln_score[gene_id] + int(l.split()[2])
    os.remove(fa1_filename)
    if not fa2_filename == fa1_filename:
        os.remove(fa2_filename)
    return(aln_score)



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import shutil
from pathlib import Path
from ete3 import NCBITaxa
from Bio import SeqIO
import time

import fdog.libs.zzz as general_fn
import fdog.libs.fasta as fasta_fn
import fdog.libs.hmm as hmm_fn
import fdog.libs.alignment as align_fn
import fdog.libs.tree as tree_fn
import fdog.libs.fas as fas_fn
import fdog.libs.output as output_fn
import fdog.libs.orthosearch as ortho_fn


##### FUNCTIONS RELATED TO CORE COMPILATION #####

def get_core_taxa_ids(coreTaxa, corepath):
    """ Get taxonomy IDs for core taxa
    Either from coreTaxa_dir, or from user input list (--coreTaxa)
    Return dictionary {taxID:<TaxName>@<TaxID>@Ver}
    """
    tax_ids = {}
    if not coreTaxa == '':
        ignored_taxa = []
        if os.path.exists(os.path.abspath(coreTaxa)):
            core_taxa = general_fn.read_file(coreTaxa)
        else:
            core_taxa = coreTaxa.split(',')

        for core_taxon in core_taxa:
            if not os.path.exists(
                    os.path.abspath(
                        '%s/%s/%s.phr' % (corepath,core_taxon,core_taxon))):
                ignored_taxa.append(core_taxon)
            else:
                id = core_taxon.split('@')[1]
                if not id in tax_ids:
                    tax_ids[id] = core_taxon
        if len(ignored_taxa) > 0:
            print(
                'WARNING: %s taxa cannot be found at %s\n%s'
                % (len(ignored_taxa), corepath, ignored_taxa))
    else:
        tax_ids = general_fn.get_ids_from_folder(corepath, 'coreTaxa_dir')
    return(tax_ids)


def initiate_core_files(
        seqFile, seqName, refspec, seed_id, hmmpath, annopath, aligner, fasOff):
    hmm_dir = '%s/%s/hmm_dir' % (hmmpath, seqName)
    Path(hmm_dir).mkdir(parents = True, exist_ok = True)
    aln_file = '%s/%s/hmm_dir/%s.aln' % (hmmpath, seqName, seqName)
    aln_seed = align_fn.do_align(aligner, seqFile)
    fasta_fn.write_fasta(aln_seed, aln_file)
    hmm_file = '%s/%s/hmm_dir/%s.hmm' % (hmmpath, seqName, seqName)
    hmm_seed = hmm_fn.create_hmm(aln_file, hmm_file)

    fa_file = '%s/%s/%s.fa' % (hmmpath, seqName, seqName)
    seed_id_mod = '%s|%s|%s' % (seqName, refspec, seed_id)
    input_seed = SeqIO.parse(seqFile,'fasta')
    with open(fa_file, 'w') as initial_core_fa:
        for fa in input_seed:
            initial_core_fa.write('>%s\n%s\n' % (seed_id_mod, str(fa.seq)))

    seed_json = ''
    if not fasOff == True:
        seed_json = fas_fn.get_anno_fas(
            seqName, refspec, seed_id, str(fa.seq), hmmpath, annopath)
    return(aln_file, fa_file, hmm_file, seed_json)


def store_cand_reults(args):
    """ Save intermediate results for a candidate ortholog
    Including:
    1) Candidate joined score of fas & normalised aln score
    in dictionary {taxID:score}
    2) Candidate fasta sequence in dictionary {taxID:fasta_seq}
    3) Update current (best) candidate, normalised aln score and joined score
    """
    (cand_taxid, cand_score, cand_seq,
        curr_cand, curr_aln_score, aln_score_normalized,
        fas_score, fas_dict, ortho_id, ortho_seq) = args
    cand_score[cand_taxid] = float(fas_score) + float(aln_score_normalized)
    cand_seq[cand_taxid] = {ortho_id:ortho_seq}
    curr_aln_score = aln_score_normalized
    curr_candi_score = float(fas_score) + float(aln_score_normalized)
    curr_cand = cand_taxid
    return(
        cand_score, cand_seq, curr_cand, curr_aln_score,
        curr_candi_score, fas_dict)


def validate_candidate(args):
    """ Validate candidate based on its normalised aln score and fas score """
    (aln_score_normalized, cand_args, calc_fas_args, variable_args, debugCore, distDeviation) = args
    (cand_score, cand_seq, curr_cand, curr_aln_score,
                        curr_candi_score, fas_dict) = variable_args
    (cand_taxid, ortho_id, ortho_seq, next_node, first_cand) = cand_args
    (fasOff, seqName, seed_json, spec, seq_id, seq, hmmpath, annopath) = calc_fas_args

    if first_cand == True:
        threshold = 0
    else:
        if next_node == True:
            threshold = curr_candi_score * (1 + distDeviation)
            type = ' (STRICT) '
        else:
            threshold = curr_candi_score
            type = ''

    if aln_score_normalized > threshold - 1:
        if not '%s_%s' % (spec, seq_id) in fas_dict:
            fas_score = fas_fn.calc_fas_cand(calc_fas_args)
            fas_dict['%s_%s' % (spec, seq_id)] = fas_score
            output_fn.print_debug(
                debugCore, '', '-FAS: %s' % fas_score)
        else:
            fas_score = fas_dict['%s_%s' % (spec, seq_id)]
        if float(fas_score) + float(aln_score_normalized) > threshold:
            variable_args = store_cand_reults(
                        [cand_taxid,
                        cand_score, cand_seq,
                        curr_cand, curr_aln_score,
                        aln_score_normalized, fas_score, fas_dict,
                        ortho_id, ortho_seq])
        else:
            output_fn.print_debug(
                debugCore, '',
                '-Joined score not higher than the prev%s! Skip...' % type)
    else:
        output_fn.print_debug(
            debugCore, '',
            '-Aln score %s not higher%s! Skip...' % (aln_score_normalized, type))
    return(variable_args)


def compile_core(args):
    """ Core compilation """
    (seqFile, seqName, refspec, seed_id, coreArgs, pathArgs, orthoArgs, otherArgs, debug) = args
    (minDist, maxDist, coreSize, coreTaxa, distDeviation, alnStrategy, fasOff) = coreArgs
    (outpath, hmmpath, corepath, searchpath, annopath) = pathArgs
    (cpus, debugCore, silentOff, noCleanup, force, append) = otherArgs
    aligner = orthoArgs[-1]
    otherArgs.insert(0, 'NA')

    ncbi = NCBITaxa()
    ### get taxonomy lineage of refspec
    refspec_id = refspec.split('@')[1]
    refspec_lineage = ncbi.get_lineage(refspec_id)

    ### get rank ID and its index in the refspec lineage
    (min_rank, max_rank) = tree_fn.get_rank_range(refspec_lineage, minDist, maxDist, ncbi)
    output_fn.print_debug(debugCore, 'Min & Max-rank', '%s\t%s' % (min_rank, max_rank))

    ### create taxonomy tree from list of core tax
    tax_ids = get_core_taxa_ids(coreTaxa, corepath)
    tree = ncbi.get_topology(tax_ids.keys(), intermediate_nodes = True)
    if debugCore:
        print(tree)

    ### INITIATE FA, ALN, HMM [and anno FAS] FILE FOR SEED
    (aln_file, fa_file, hmm_file, seed_json) = initiate_core_files(
        seqFile, seqName, refspec, seed_id, hmmpath, annopath, aligner, fasOff)

    ### get list of taxa within min and max rank of refspec
    node_dict = tree_fn.get_leaves_dict(
                    refspec_lineage, tree,
                    list(min_rank.values())[0], list(max_rank.values())[0])
    output_fn.print_debug(debugCore, 'Node dictionary', node_dict)

    ### traverse the core taxa tree
    added_taxa = {}
    ignored_taxa = []
    fas_dict = {}
    previous_added_taxon = refspec_id
    for round in range(coreSize - 1):
        output_fn.print_stdout(silentOff, '---------- ROUND %s ----------' % round)
        output_fn.print_debug(
            debugCore, 'CORE COMPILATION',
            '---------- ROUND %s ----------' % round)
        aln_scores = align_fn.calc_aln_score(fa_file, fa_file, alnStrategy, debugCore)
        max_aln_score = max(aln_scores.values()) #0
        if max_aln_score == 0:
            exit('ERROR: Something went wrong with FASTA36. Please run debugCore to investigate!')
        flag_round = 0 # use to stop current round if an ortholog was added
        output_fn.print_debug(debugCore, '', 'ADDED TAXA: %s' % added_taxa.keys())
        cand_seq = {}
        cand_score = {}
        curr_cand = ''
        curr_aln_score = 0
        curr_candi_score = 0
        next_node = False
        for node_id, leaves in node_dict.items():
            if flag_round == 1:
                break
            output_fn.print_debug(
                debugCore, '',
                'NODE %s - %s' % (node_id, ncbi.get_rank([node_id])))
            output_fn.print_debug(
                debugCore, '',
                '-MAX ALN SCORE: %s' % max_aln_score)
            output_fn.print_debug(
                debugCore, '',
                '-PREVIOUS ADDED: %s' % previous_added_taxon)
            leaves.reverse()
            flag_node = 0
            for leaf in leaves:
                if not leaf in tax_ids:
                    continue
                if flag_node == 1:
                    break
                if not leaf == refspec_id and \
                        not leaf in added_taxa and \
                        not leaf in ignored_taxa:
                    output_fn.print_debug(debugCore, '','')
                    output_fn.print_debug(
                        debugCore, '',
                        'Leaf %s - %s' % (leaf, tax_ids[leaf]))
                    if len(curr_cand) > 0 and \
                            not curr_cand in node_dict[node_id]:
                        next_node = True
                        output_fn.print_debug(
                            debugCore, '',
                            '-Current_candidate from different node: %s' \
                            % curr_cand)
                        if curr_candi_score * (1 + distDeviation) > 2:
                            output_fn.print_debug(
                                debugCore, '',
                                '# Current score cannot be defeater! Stop this node!')
                            break
                    else:
                        next_node = False
                        output_fn.print_debug(
                            debugCore, '',
                            '-Current_candidate: %s' % curr_cand)
                    output_fn.print_debug(
                        debugCore, '',
                        '-Current_aln_score: %s' % curr_aln_score)
                    output_fn.print_debug(
                        debugCore, '',
                        '-Current_candi_score: %s' % curr_candi_score)
                    ### compare taxonomy rank with previous added taxon
                    ### ignore if this leaf closer to the refspec than to
                    ### previous added taxon
                    ancestor_to_ref = tree_fn.get_ancestor(refspec_id, leaf, ncbi)
                    check_ancestor_to_ref = tree_fn.check_common_ancestor(
                            previous_added_taxon, list(ancestor_to_ref.keys())[0],
                            minDist, maxDist, ncbi)
                    if check_ancestor_to_ref == 0:
                        ignored_taxa.append(leaf)
                        output_fn.print_debug(
                            debugCore, '',
                            '-Closer to refspec than previous added taxon!')
                        continue
                    ### continue process only if this leaf is within min and max rank
                    ### of the previous added taxon
                    ancestor = tree_fn.get_ancestor(previous_added_taxon, leaf, ncbi)
                    check_ancestor = tree_fn.check_common_ancestor(
                            previous_added_taxon, list(ancestor.keys())[0],
                            minDist, maxDist, ncbi)
                    if check_ancestor == 1:
                        output_fn.print_debug(
                            debugCore, '',
                            '-Ancestor %s with %s accepted' \
                            % (ancestor, previous_added_taxon))
                        ### run ortholog search
                        otherArgs[2] = debug
                        otherArgs[0] = tax_ids[leaf]
                        hamstr_out = ortho_fn.run_hamstr([seqName, refspec, pathArgs,
                            orthoArgs, otherArgs])
                        if len(hamstr_out) > 1:
                            ### calculate alignment score
                            ortho = list(hamstr_out.items())[-1]
                            tmp_fa = '%s/%s/%s_%s.fa' \
                                % (hmmpath, seqName, seqName, leaf)
                            with open(tmp_fa, 'w') as tmp_fa_out:
                                tmp_fa_out.write('>%s\n%s\n' \
                                    % (ortho[0][0:len(ortho[0])-2], ortho[1]))
                            aln_score = align_fn.calc_aln_score(fa_file, tmp_fa, alnStrategy, debugCore)
                            output_fn.print_debug(
                                debugCore, '',
                                '-Max: %s - Aln: %s' % (max_aln_score, aln_score))
                            aln_score_normalized = \
                                    list(aln_score.values())[0] / max_aln_score
                            output_fn.print_debug(
                                debugCore, '',
                                '-Normalized_aln_score: %s' % aln_score_normalized)
                            os.remove(tmp_fa)

                            ### validate candidate
                            if len(cand_score) == 0 \
                                    and len(curr_cand) == 0:
                                first_cand = True
                            else:
                                first_cand = False
                            calc_fas_args = (fasOff, seqName, seed_json,
                                tax_ids[leaf], ortho[0].split('|')[-2],
                                ortho[1] ,hmmpath, annopath)
                            cand_args = (leaf, ortho[0][0:len(ortho[0])-2],
                                ortho[1], next_node, first_cand)
                            variable_args = (cand_score, cand_seq,
                                curr_cand, curr_aln_score,
                                curr_candi_score, fas_dict)
                            (cand_score, cand_seq, curr_cand,
                                curr_aln_score, curr_candi_score,
                                fas_dict) = validate_candidate([
                                        aln_score_normalized, cand_args,
                                        calc_fas_args, variable_args, debugCore,
                                        distDeviation])
                            if curr_candi_score == 2:
                                flag_node = 1
                                output_fn.print_debug(
                                    debugCore, '',
                                    '-Max score achieved! Stop this node!')
                        elif not len(hamstr_out) > 1 and len(cand_seq) == 0:
                            ignored_taxa.append(leaf)
                            output_fn.print_debug(
                                debugCore, '',
                                '-No ortholog found!')
                    else:
                        ignored_taxa.append(leaf)
                        output_fn.print_debug(
                            debugCore, '',
                            '-Not considered due to ancestor %s with %s\n' \
                                % (ancestor, previous_added_taxon))
                else:
                    output_fn.print_debug(
                        debugCore, '',
                        '%s - %s skipped' % (leaf, tax_ids[leaf]))
            output_fn.print_debug(
                debugCore, '', 'Node candidates: %s' % cand_score)
            if len(cand_score) > 0 \
                    and cand_score[curr_cand] == 2:
                output_fn.print_debug(
                    debugCore, '',
                    '# MAX SCORE ACCHIEVED! Stop this round!')
                flag_round = 1
            if next_node == True \
                    and cand_score[curr_cand] * (1 + distDeviation) > 2:
                output_fn.print_debug(
                    debugCore, '',
                    '# CURRENT SCORE CANNOT BE DEFEATED! Stop this round!')
                flag_round = 1

        if len(cand_seq) > 0:
            output_fn.print_debug(
                debugCore, '',
                '# ADD THIS TAXON TO CORE GROUP\t%s - %s\n' \
                % (curr_cand, tax_ids[curr_cand]))
            previous_added_taxon = curr_cand
            added_taxa[curr_cand] = {tax_ids[curr_cand]:cand_score[curr_cand]}
            ### update seqName.fa and hmm_dir/seqName.hmm
            fasta_fn.append_to_fasta_file(fa_file, cand_seq[curr_cand])
            aln_seed = align_fn.do_align(aligner, fa_file)
            fasta_fn.write_fasta(aln_seed, aln_file)
            hmm_seed = hmm_fn.create_hmm(aln_file, hmm_file)
            os.remove(aln_file)
    ### remove temp json files
    for file in os.listdir('%s/%s' % (hmmpath, seqName)):
        if file.endswith('.json'):
            os.remove('%s/%s/%s' % (hmmpath, seqName, file))
    output_fn.print_debug(
        debugCore, 'CORE COMPILATION',
        'All added taxa %s' % added_taxa)
    if len(added_taxa) < coreSize - 1:
        output_fn.print_stdout(
            silentOff,
            'WARNING: Only %s/%s orthologs in the core group' \
            % (len(added_taxa) + 1, coreSize))


def run_compile_core(args):
    (seqFile, seqName, refspec, seed_id, reuseCore, forceCore, coreArgs,
            pathArgs, orthoCoreArgs, otherCoreArgs, debug) = args
    (outpath, hmmpath, corepath, searchpath, annopath) = pathArgs
    (cpus, debugCore, silentOff, noCleanup, force, append) = otherCoreArgs[-6:]
    begin = time.time()
    fdogPath = os.path.realpath(__file__).replace('/libs/corecompile.py','')
    align_fn.check_fasta36_executable(fdogPath)

    coreHmmfile = '%s/%s/hmm_dir/%s.hmm' % (hmmpath, seqName, seqName)
    coreHmmfile = os.path.abspath(coreHmmfile)
    compile_core_check = 1
    ncbi = ''
    if reuseCore == True:
        general_fn.check_file_exist(coreHmmfile)
        compile_core_check = 0
    else:
        if os.path.exists(coreHmmfile):
            if forceCore == True:
                print('WARNING: Existing %s core group will be deleted!' % seqName)
                shutil.rmtree('%s/%s' % (hmmpath, seqName))
            else:
                sys.exit(
                    'WARNING: Core group %s exists in %s! ' % (seqName, hmmpath)
                    + 'You still can run with --forceCore or --reuseCore option')
    if compile_core_check == 1:
        compile_core([seqFile, seqName, refspec, seed_id, coreArgs, pathArgs,
                    orthoCoreArgs, otherCoreArgs[-6:], debug])
    end = time.time()
    return([seqName, '{:5.3f}s'.format(end - begin)])



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
from pathlib import Path
from Bio import SeqIO
import subprocess
from ete3 import NCBITaxa
import re
from datetime import datetime
from collections import OrderedDict

import fdog.libs.zzz as general_fn
import fdog.libs.blast as blast_fn
import fdog.libs.fasta as fasta_fn
import fdog.libs.tree as tree_fn

##### FUNCTIONS RELATED TO ADDING NEW TAXON TO FDOG DATABASE #####

def check_conflict_opts(replace, delete):
    """ Check if both replace and delete option are specified """
    if delete:
        if replace:
            sys.exit('*** ERROR: only one option can be choose between "--replace" and "--delete"')
    if replace:
        if delete:
            sys.exit('*** ERROR: only one option can be choose between "--replace" and "--delete"')


def get_paths(outPath, fdogPath, searchpath, corepath, annopath):
    """ Get path to searchTaxa_dir, coreTaxa_dir and annotation_dir """
    if outPath == '':
        pathconfigFile = fdogPath + '/bin/pathconfig.yml'
        if not os.path.exists(pathconfigFile):
            sys.exit('No pathconfig.yml found. Please run fdog.setup (https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).')
        cfg = general_fn.load_config(pathconfigFile)
        try:
            outPath = cfg['dataPath']
        except:
            try:
                corepath = cfg['corepath']
            except:
                pass
            try:
                searchpath = cfg['searchpath']
            except:
                pass
            try:
                annopath = cfg['annopath']
            except:
                pass

    outPath = os.path.abspath(outPath)
    if not searchpath:
        searchpath = outPath + '/searchTaxa_dir/'
    searchpath = os.path.abspath(searchpath)
    if not corepath:
        corepath = outPath + '/coreTaxa_dir/'
    corepath = os.path.abspath(corepath)
    if not annopath:
        annopath = outPath + '/annotation_dir/'
    annopath = os.path.abspath(annopath)
    return(outPath, searchpath, corepath, annopath)

def create_folders(searchpath, corepath, annopath, spec_name, coreTaxa, noAnno):
    """ Create searchTaxa_dir, coreTaxa_dir and annotation_dir in output folder """
    Path(searchpath).mkdir(parents = True, exist_ok = True)
    genome_path = '%s/%s' %  (searchpath, spec_name)
    Path(genome_path).mkdir(parents = True, exist_ok = True)
    if coreTaxa:
        Path(corepath).mkdir(parents = True, exist_ok = True)
    if not noAnno:
        Path(annopath).mkdir(parents = True, exist_ok = True)
    return(genome_path)


def generate_spec_name(tax_id, name, ver):
    """ Create species name with the format <ABBR_NAME>@<NCBI_TAXID>@<VER> """
    if name == "":
        ncbi_name = tree_fn.check_tax_id(tax_id)
        name = tree_fn.abbr_ncbi_name(ncbi_name)
    return(name+'@'+tax_id+'@'+ver)


def create_genome(args):
    """ Create fa and fai in searchTaxa_dir """
    (faIn, genome_path, spec_name, force, replace, delete) = args
    ### load fasta seq
    in_seq = SeqIO.to_dict((SeqIO.parse(open(faIn), 'fasta')))
    if not os.path.exists(genome_path):
        Path(genome_path).mkdir(parents = True, exist_ok = True)
    genome_file = '%s/%s.fa' % (genome_path, spec_name)
    if (not os.path.exists(os.path.abspath(genome_file))) or (os.stat(genome_file).st_size == 0) or force:
        f = open(genome_file, 'w')
        pipe = 0
        long_id = 0
        mod_id_index = 0
        id_dict = {} # id_dict[ori_id] = mod_id
        for id in in_seq:
            ori_id = id
            seq = str(in_seq[id].seq)
            ### check if ID contains empty char or pipe
            if ' ' in id:
                sys.exit('\033[91mERROR: Sequence IDs (e.g. %s) must not contain space(s)!\033[0m' % id)
            else:
                if '|' in id:
                    tmp = re.split('[_|]', id)
                    tmp = list(OrderedDict.fromkeys(tmp))
                    pipe = 1
                    id = '_'.join(tmp)
                    if not ori_id in id_dict:
                        id_dict[ori_id] = id
            ### check if id longer than 20 character
            if len(id) > 20:
                long_id = 1
                mod_id_index = mod_id_index + 1
                id = '%s_%s' % (spec_name.split('@')[1], mod_id_index)
                id_dict[ori_id] = id
            ### check if seq contains special characters
            if seq[-1] == '*':
                seq = seq[:-1]
            specialChr = 'no'
            if any(c for c in seq if not c.isalpha()):
                specialChr = 'yes'
            if specialChr == 'yes':
                if replace or delete:
                    if replace:
                        seq = re.sub('[^a-zA-Z]', 'X', seq)
                    if delete:
                        seq = re.sub('[^a-zA-Z]', '', seq)
                else:
                    sys.exit('\033[91mERROR: %s sequence contains special character!\033[0m\nYou can use --replace or --delete to solve it.' % (id))
            f.write('>%s\n%s\n' % (id, seq))
        f.close()
        ### create index file
        fasta_fn.read_fasta(genome_file)
        ### write .checked file
        cf = open(genome_file+'.checked', 'w')
        cf.write(str(datetime.now()))
        cf.close()
        ### write ID mapping file and give warning if ID changed
        if len(id_dict) > 0:
            mapping_file = '%s.mapping' % genome_file
            with open(mapping_file, 'w') as mp:
                for id in in_seq:
                    if id in id_dict:
                        mp.write(f'{id}\t{id_dict[id]}\n')
                    else:
                        mp.write(f'{id}\t{id}\n')
            if pipe == 1:
                print('\033[94mWARNING: Sequence IDs contain pipe(s). They will be replaced by "_"!\033[0m')
            if long_id == 'yes':
                print('\033[94mWARNING: Some headers longer than 80 characters have been automatically shortened.\033[0m')
            print('\033[94mPlease check the %s file for details!\033[0m' % mapping_file)
    else:
        print(genome_path + '/' + spec_name + '.fa already exists!')
    return(genome_file)


def create_blastdb(args):
    """ Create blastdb for a given fasta genome_file """
    (searchpath, corepath, outPath, spec_name, genome_file, force, silent) = args
    blast_path = '%s/%s' % (corepath, spec_name)
    if (not os.path.exists(os.path.abspath('%s/%s.phr' % (blast_path, spec_name)))) or force:
        blast_fn.make_blastdb([spec_name, genome_file, outPath, corepath, searchpath, silent])
        ### make symlink to fasta files
        fa_in_genome = "%s/%s/%s.fa" % (searchpath, spec_name, spec_name)
        fai_in_genome = "%s/%s/%s.fa.fai" % (searchpath, spec_name, spec_name)
        fa_in_blast = "%s/%s.fa" % (blast_path, spec_name)
        fai_in_blast = "%s/%s.fa.fai" % (blast_path, spec_name)
        if not os.path.exists(fa_in_blast):
            os.symlink(fa_in_genome, fa_in_blast)
        if not os.path.exists(fai_in_blast):
            os.symlink(fai_in_genome, fai_in_blast)
    else:
        print('Blast DB already exists!')


def create_annoFile(annopath, genome_file, cpus, force):
    """ Create annotation json for a given genome_file """
    annoCmd = 'fas.doAnno -i %s -o %s --cpus %s' % (genome_file, annopath, cpus)
    if force:
        annoCmd = annoCmd + " --force"
    try:
        subprocess.call([annoCmd], shell = True)
    except:
        print('\033[91mERROR: Problem with running fas.doAnno. You can check it with this command:\n%s\033[0m' % annoCmd)



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This script is used to set default data directories for fdog.
#  These include the path to the core taxa (coreTaxa_dir),
#  search taxa (searchTaxa_dir) and FAS annotation json files (annotation_dir).
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import os
import argparse
from importlib.metadata import version, PackageNotFoundError

import fdog.libs.zzz as general_fn
import fdog.checkData as check_data_fn


def set_data_path(searchpath, corepath, annopath, checkOff):
    """ Set default fDOG data path to pathconfig.yml file """
    fdogPath = os.path.realpath(__file__).replace('/setPaths.py','')
    pathconfigFile = fdogPath + '/bin/pathconfig.yml'
    flag = 0
    if os.path.exists(pathconfigFile):
        print('Default fDOG data paths in %s will be overwritten! Enter to continue.' % pathconfigFile)
        if general_fn.query_yes_no(''):
            flag = 1
    else:
        flag = 0

    if not checkOff:
        caution = check_data(searchpath, corepath, annopath)
        if caution == 1:
            print('Check done! Data are ready to use WITH CAUTION! Are you sure to add these paths as default? (Y/N)')
            if general_fn.query_yes_no('', default='no'):
                flag = 1
            else:
                flag = 0
    else:
        print('WARNING: Data will not be checked! Run fdog.checkData if you encounter any problems!')

    if flag == 1:
        with open(pathconfigFile, 'w') as cf:
            cf.write('corepath: \'%s\'\n' % corepath)
            cf.write('searchpath: \'%s\'\n' % searchpath)
            cf.write('annopath: \'%s\'\n' % annopath)
        print('Finished! New data paths have been saved in %s' % pathconfigFile)
    else:
        print(f'{pathconfigFile} remains unchanged!')


def check_data(searchpath, corepath, annopath):
    """ Perform data check """
    caution = check_data_fn.run_check([searchpath, corepath, annopath, False, False, False, False, False])
    return(caution)


def main():
    fdog_version = version("fdog")
    parser = argparse.ArgumentParser(description='You are running fDOG version ' + str(fdog_version) + '.')
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument('--searchpath', help='Path to search taxa folder (e.g. fdog_data/searchTaxa_dir)', action='store', default='', required=True)
    required.add_argument('--corepath', help='Path to core taxa folder (e.g. fdog_data/coreTaxa_dir)', action='store', default='', required=True)
    required.add_argument('--annopath', help='Path to annotation folder (e.g. fdog_data/annotation_dir)', action='store', default='', required=True)
    optional.add_argument('--checkOff', help='Turn off checking for valid data', action='store_true', default=False)

    args = parser.parse_args()
    searchpath = os.path.abspath(args.searchpath)
    corepath = os.path.abspath(args.corepath)
    annopath = os.path.abspath(args.annopath)
    checkOff = args.checkOff

    set_data_path(searchpath, corepath, annopath, checkOff)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This script is used to list all available taxa of the installed fdog
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
from ete3 import NCBITaxa

import fdog.libs.zzz as general_fn

def getNcbiName(taxonName):
    ncbi = NCBITaxa()
    taxId = taxonName.split('@')[1]
    try:
        name = ncbi.get_taxid_translator([taxId])[int(taxId)]
    except:
        name = taxonName
    return(name)


def getTaxa():
    # get data path
    fdogPath = os.path.realpath(__file__).replace('/showTaxa.py','')
    pathconfigFile = fdogPath + '/bin/pathconfig.yml'
    if not os.path.exists(pathconfigFile):
        sys.exit('No pathconfig.yml found. Please run fdog.setup (https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).')
    with open(pathconfigFile) as f:
        cfg = general_fn.load_config(pathconfigFile)
        try:
            dataPath = cfg['datapath']
        except:
            dataPath = 'several places!'
        try:
            corepath = cfg['corepath']
        except:
            corepath = dataPath + '/coreTaxa_dir'
        general_fn.check_file_exist(corepath)
        try:
            searchpath = cfg['searchpath']
        except:
            searchpath = dataPath + '/searchTaxa_dir'
        general_fn.check_file_exist(searchpath)

    # print taxa in coreTaxa_dir
    print('##### Data found at %s' % dataPath)
    print('\n##### Taxa in the core sets, which can be used as reference species #####\n')
    for taxon in sorted(os.listdir(corepath)):
        if os.path.isdir(f'{corepath}/{taxon}'):
            print('%s\t%s' % (taxon, getNcbiName(taxon)))

    # print taxa in searchTaxa_dir
    print('\n##### Search taxa. in which you can search orthologs #####\n')
    for taxon in sorted(os.listdir(searchpath)):
        if os.path.isdir(f'{searchpath}/{taxon}'):
            print('%s\t%s' % (taxon, getNcbiName(taxon)))


def main():
    getTaxa()

if __name__ == '__main__':
    main()
